﻿/**************************************************************************
 * XYZ.cs
 * xyz表現の座標のための構造体 for C#
 * 
 * 開発者　：森下功啓（K.Morhista Kumamoto-University）
 * 開発履歴：2011/7/3   整備開始
 *           2011/7/6   2地点間の距離を求める、GetDistanceのデバッグ。単位長までしか計算していなかった…。
 *           2011/7/17  XYZとBLHに二項演算子を追加
 *           2011/9/19  Open(string fname)のバグを修正
 *                      GetPositions(),GetPositioningResults()において、ubloxのNMEAフォーマットに対応（現時点での要求仕様を満たすだけなので万能ではないことに注意）
 *                      PositioningInfosのToString()に高度も含まれるように変更
 *           2011/11/11 よく分かっていないが、シリアル化した方が良さそうとのことで[Serializable]を片っ端からつけてみた。今後どうなるかは・・・よく分からない。
 *           2011/11/20 BLH構造体にToString()を追加
 *           2011/12/2  XML構文エラーがある部分を訂正した
 *           2011/12/15 BLH構造体にToString()を追加
 *           2012/4/22  コメントを少しだけ付け直した。
 *                      NMEAクラスは現時点ではバイナリの処理が不可能だが、そこんとこを何とかしたい。
 *           2012/4/24  ソースコードを分割した。
 * ***********************************************************************/
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using GNSS.GeodeticDatum;

namespace GNSS
{
    /*2011/7/6 動作確認済み*/
    /// <summary>
    /// x,y,zを格納する構造体
    /// <para>
    /// 半径の取得も簡単
    /// BLH座標系（緯度・経度・楕円体高[m]）への変換も簡単
    /// 座標変換処理等を実装する必要もあるのだろうけど、現時点ではWGS84専用となっている。
    /// </para>
    /// </summary>
    [Serializable]
    public struct XYZ
    {
        /************************* メンバ変数 ***************************/
        /// <summary>x座標[m]</summary>
        public double x;
        /// <summary>y座標[m]</summary>
        public double y;
        /// <summary>z座標[m]</summary>
        public double z;
        /******************** 演算子 ********************************/
        /// <summary>
        /// 二項+演算子（これで足し算が簡単にできる）
        /// </summary>
        /// <param name="c1">被加算値</param>
        /// <param name="c2">加算値</param>
        /// <returns>2値を加算した結果</returns>
        public static XYZ operator +(XYZ c1, XYZ c2)
        {
            return new XYZ(c1.x + c2.x, c1.y + c2.y, c1.z + c1.z);
        }
        /// <summary>
        /// 二項-演算子（これで足し算が簡単にできる）
        /// </summary>
        /// <param name="c1">被減算値</param>
        /// <param name="c2">減算値</param>
        /// <returns>2値の引き算の結果</returns>
        public static XYZ operator -(XYZ c1, XYZ c2)
        {
            return new XYZ(c1.x - c2.x, c1.y - c2.y, c1.z - c1.z);
        }
        /******************** プロパティ ****************************/
        /// <summary>
        /// 半径（地心からの距離）
        /// </summary>
        public double r 
        { 
            get {
                return (Math.Sqrt(Math.Pow(this.x, 2) + Math.Pow(this.y, 2) + Math.Pow(this.z, 2))); 
            } 
        }
        /******************** メソッド ****************************/
        /// <summary>
        /// XYZからBLH座標系（緯度・経度・楕円体高[m]）へ変換
        /// </summary>
        /// <returns>BLHに変換した結果</returns>
        public BLH ToBLH()
        {
            BLH ans = new BLH();
            double a, b, e, n, h, p, t, sint, cost;

            ans.H = -WGS84.a;
            if (this.x == 0 && this.y == 0 && this.z == 0) return ans;

            a = WGS84.a;                        // 長半径
            b = WGS84.b;                        // 短半径
            e = Math.Sqrt(WGS84.e2);            // 離心率

            //　座標変換のためのパラメータ
            h = Math.Pow(a, 2) - Math.Pow(b, 2);
            p = Math.Sqrt(Math.Pow(this.x, 2) + Math.Pow(this.y, 2));
            t = Math.Atan2(this.z * a, p * b);
            sint = Math.Sin(t);
            cost = Math.Cos(t);

            ans.B = Math.Atan2(this.z + h / b * Math.Pow(sint, 3), p - h / a * Math.Pow(cost, 3));  // 緯度[rad]を計算する
            ans.L = Math.Atan2(this.y, this.x);                                                     // 経度[rad]を求める
            n = a / Math.Sqrt(1 - WGS84.e2 * Math.Pow(Math.Sin(ans.B), 2));                         // 卯酉線曲率半径
            ans.H = (p / Math.Cos(ans.B)) - n;                                                      // 楕円体高[m]
            return ans;
        }
        /// <summary>
        /// 文字列化
        /// </summary>
        /// <returns>メンバ変数x,y,zを文字列化したしたstring型変数</returns>
        public override string ToString()
        {
            StringBuilder sb = new StringBuilder(50);                      // 予め大きなメモリ容量を確保しておく

            sb.Append(this.x.ToString("0.0")).Append(",").Append(this.y.ToString("0.0")).Append(",").Append(this.z.ToString("0.0"));
            return sb.ToString();
        }
        /// <summary>
        /// コンストラクタ
        /// </summary>
        public XYZ(double _x = 0.0d, double _y = 0.0d, double _z = 0.0d) { this.x = _x; this.y = _y; this.z = _z; }      // 初めから値を指定していてもよい
    }
}
